package edu.cmu.cs.ls.keymaerax.btactics.arithmetic

import edu.cmu.cs.ls.keymaerax.core._

object VirtualSubstitution {
  val epsilon : Variable = Variable("eps")
  val infinity : Variable = Variable("inf")


  /** Remove minus take in an FOLR term P and converts statements of the form (a - b) into a + (-b), return a term with no Minus Constructors,
    * only Negs
    */
  // List[Variable] Set[Variable]
  def RemoveMinus(input : Term, free : Variable):Term =  {
    input match {
      case Plus(a,b) => Plus(RemoveMinus(a,free),RemoveMinus(b,free))
      case Times(a,b) => Times(RemoveMinus(a,free),RemoveMinus(b,free))
      case Minus(a,b) => Plus(RemoveMinus(a,free),Neg(RemoveMinus(b,free)))
      case Power(a,Number(s)) => Power(RemoveMinus(a,free),Number(s))
      case Neg(a) => Neg(RemoveMinus(a,free))
      case Divide(a,b) => Divide(RemoveMinus(a,free),RemoveMinus(b,free))
      case _ => input
    }
  }

  /** *
    * Normalize Poly will take in a FOLR term without Minus constructors and a variable "x" and normalize to be a poylnomial to be a sum of terms of the form
    *  a*x^k.
    */

  def NormalizePoly(input : Term, free :  Variable) :Term = {
    RemoveMinus(input,free) match {
      /** check if it's x */
      case freevar => Times(Number(1),Power(free,Number(1)))
      /** check if it's (-P) */
      case Neg(a) =>
        val minusOne = Times(Number(~1),Power(free,Number(0)))
        MultNormalizedPolys(minusOne,NormalizePoly(a,free),free)
      /** check if it's x^k, we assume the exponents are non-negative integers */
      case Power(freevar,Number(s)) => input
      /** check if it's a + b */
      case Plus(a,b) => AddNormalizedPolys(NormalizePoly(a,free),NormalizePoly(b,free),free)
      /** check if it's a*b */
      case Times(a,b) => MultNormalizedPolys(NormalizePoly(a,free),NormalizePoly(b,free),free)
      /** check if it's a^k, we assume the exponents are non-negative integers */
      case Power(a,Number(s)) =>
        if(s > 0){
          MultNormalizedPolys(a,NormalizePoly(Power(a,Number(s-1)),free),free)
        }
        else {
          Times(Number(1),Power(free,Number(0)))
        }
      /** check if it's a/b, assuming b is x free */
      case Divide(a,b) => MultNormalizedPolys(Divide(Number(1),NormalizePoly(b,free)),NormalizePoly(a,free),free)
      /** if no other case, it must be of the form b, where b is a symbol not of x*/
      case _ => Times(input,Power(free,Number(0)))
    }
  }
  def AddNormalizedPolys(poly1 : Term, poly2 : Term, freevar :  Variable) : Term = {
    (poly1,poly2) match {
      /** In the first case, we consider an input of the form a*x^n + b*x^m, where a,b are x free */
      case (Times(a1,Power(freevar1,Number(n1))),Times(a2,Power(freevar2,Number(n2)))) =>
        if(n1 > n2){
          Plus(Times(a1,Power(freevar,Number(n1))),Times(a2,Power(freevar,Number(n2))))
        }
        else
        if(n1 < n2){
          Plus(Times(a2,Power(freevar,Number(n2))),Times(a1,Power(freevar,Number(n1))))
        }
        else{
          Times(Plus(a1,a2),Power(freevar,Number(n1)))
        }

      /** In the second case, we consider an input of the form a*x^n + ((b_n)x^(m_n) + ... + (b_0)x^0),
        * where a and the bs are x free */
      case (Times(a1,Power(freevar1,Number(n1))),Plus(Times(a2,Power(freevar2,Number(n2))),b2)) =>
        if(n1 > n2){
          Plus(Times(a1,Power(freevar,Number(n1))),poly2)
        }
        else
        if(n1 < n2){
          Plus(Times(a2,Power(freevar,Number(n2))),AddNormalizedPolys(poly1,b2,freevar))
        }
        else{
          Plus(Times(Plus(a1,a2),Power(freevar,Number(n1))),b2)
        }

      /** In the third case, we consider an input of the form ((a_s)x^(n_s) + ... + (a_0)x^0) + bx^n
        * where the as and b are x free */
      case(Plus(Times(a1,Power(freevar1,Number(n1))),b1),Times(a2,Power(freevar2,Number(n2))))
        if freevar1 == freevar && freevar2 == freevar =>
        if(n1 > n2){
          Plus(Times(a1,Power(freevar,Number(n1))),AddNormalizedPolys(b1,poly2,freevar))
        }
        else
        if(n1 < n2){
          Plus(Times(a2,Power(freevar,Number(n2))),poly1)
        }
        else{
          Plus(Times(Plus(a1,a2),Power(freevar,Number(n1))),b1)
        }

      /** In the fourth case, we consider an input of the form ((a_s)x^(n_s) + ... + (a_0)x^0) + ((b_n)x^(m_n) + ... +
        *  (b_0)x^0), where the as the bs are x free */
      case (Plus(Times(a1,Power(freevar1,Number(n1))),b1),Plus(Times(a2,Power(freevar2,Number(n2))),b2))
        if (freevar1 == freevar && freevar2 == freevar) =>
        if(n1 > n2){
          Plus(Times(a1,Power(freevar,Number(n1))),AddNormalizedPolys(b1,poly2,freevar))
        }
        else
        if(n1 < n2){
          Plus(Times(a2,Power(freevar,Number(n2))),AddNormalizedPolys(poly1,b2,freevar))
        }
        else{
          Plus(Times(Plus(a1,a2),Power(freevar,Number(n1))),AddNormalizedPolys(b1,b2,freevar))
        }
    }
  }
  def MultNormalizedPolys(poly1 : Term, poly2 : Term, freevar :  Variable) : Term = {
    (poly1,poly2) match {

      /** In the first case, we consider an input of the form a*x^n * b*x^m, where a and b are x free */
      case (Times(a1,Power(freevar1,Number(n1))),Times(a2,Power(freevar2,Number(n2))))
        if freevar1 == freevar && freevar2 == freevar =>
        Times(Plus(a1,a2),Power(freevar,Number(n1+n2)))

      /** In the second case, we consider an input of the form a*x^n + ((b_n)x^(m_n) + ... + (b_0)x^0)
        * where a and the bs are x free */
      case (Times(a1,Power(freevar1,Number(n1))),Plus(Times(a2,Power(freevar2,Number(n2))),b2))
       if freevar1 == freevar && freevar2 == freevar =>
        AddNormalizedPolys(Times(Plus(a1,a2),Power(freevar,Number(n1 + n2))),MultNormalizedPolys(poly1,b2,freevar),freevar)

      /** In the third case, we consider an input of the form ((a_s)x(n_s) + ... + (a_0)x^0) + b*x^m^
        * where the as and b are x free */
      case (Plus(Times(a1,Power(freevar1,Number(n1))),b1),Times(a2,Power(freevar2,Number(n2))))
        if freevar1 == freevar && freevar2 == freevar =>
        AddNormalizedPolys(Times(Plus(a1,a2),Power(freevar,Number(n1 + n2))),MultNormalizedPolys(poly1,b1,freevar),freevar)

      /** In the fourth case, we consider an input of the form ((a_s)x^(n_s) + ... + (a_0)x^0) + ((b_n)x^(m_n) +
        *  ... + (b_0)x^0), where the as and bs are x free */
      case (Plus(Times(a1,Power(freevar1,Number(n1))),b1),Plus(Times(a2,Power(freevar2,Number(n2))),b2))
        if freevar1 == freevar && freevar2 == freevar =>
        AddNormalizedPolys(Times(Plus(a1,a2),Power(freevar,Number(n1+n2))),
          AddNormalizedPolys(MultNormalizedPolys(Times(a1,Power(freevar,Number(n1))),b2,freevar),
            AddNormalizedPolys(MultNormalizedPolys(Times(a2,Power(freevar,Number(n2))),b1,freevar),
              MultNormalizedPolys(b1,b2,freevar),freevar),freevar),freevar)
    }
  }

    /** NormalizeFormula takes in a formula and a free variable x and normalizes the expressions w.r.t. x */
    def NormalizeFormula(input : Formula, freevar :  Variable) : Formula = {
      input match {
        case       Not(f) =>   Not(NormalizeFormula(f,freevar))
        case   And(f1,f2) =>   And(NormalizeFormula(f1,freevar),NormalizeFormula(f2,freevar))
        case    Or(f1,f2) =>    Or(NormalizeFormula(f1,freevar),NormalizeFormula(f2,freevar))
        case Imply(f1,f2) => Imply(NormalizeFormula(f1,freevar),NormalizeFormula(f2,freevar))
        case Equiv(f1,f2) => Equiv(NormalizeFormula(f1,freevar),NormalizeFormula(f2,freevar))

        case  Exists(xs,f) => Exists(xs,NormalizeFormula(f,freevar))
        case  Forall(xs,f) => Exists(xs,NormalizeFormula(f,freevar))

        case Greater(t1,t2)      =>      Greater (NormalizePoly(t1,freevar),NormalizePoly(t2,freevar))
        case GreaterEqual(t1,t2) => GreaterEqual (NormalizePoly(t1,freevar),NormalizePoly(t2,freevar))
        case LessEqual(t1,t2)    =>    LessEqual (NormalizePoly(t1,freevar),NormalizePoly(t2,freevar))
        case Less(t1,t2)         =>         Less (NormalizePoly(t1,freevar),NormalizePoly(t2,freevar))
        case Equal(t1,t2)        =>        Equal (NormalizePoly(t1,freevar),NormalizePoly(t2,freevar))
        case NotEqual(t1,t2)     =>     NotEqual (NormalizePoly(t1,freevar),NormalizePoly(t2,freevar))

      }
    }
    /** Degree takes in a normalized polynomial in x and returns the degree of the polynomial */
    def Degree(polynomial : Term, free :  Variable) : BigDecimal = {
      polynomial match{
        case Plus(Times(a,Power(free,Number(s))),b) => s
        case Times(a,Power(free,Number(s))) => s
        case _ => 0
      }
    }
    /** Derivative takes in a normalized polynomial in x and returns the derivative of the polynomial w.r.t. x */
    def Derivative(polynomial : Term, free : Variable) : Term = {
      polynomial match{
        case Plus(Times(a,Power(free1,Number(s))),b)
          if(free1 == free) =>
          Plus(Times(Times(a,Number(s)),Power(free,Number(s-1))),Derivative(b,free))
        case Times(a,Power(free,Number(s))) =>
          if(s == 0){
            Number(0)
          }
          else{
            Times(Times(a,Number(s)),Power(free,Number(s-1)))
          }

      }
    }
    /** Replace takes in normalized comparison formula and x and, assuming the degree of the polynomials is at most 2 w.r.t x.
      * It returns a requirement and a list of substitutions for replacement, returning true
      * in the requirement if there are no requirements, and returning false if the terms of have degree > 2. */
    def Replacement(input : ComparisonFormula, free :  Variable) : List[(Formula, List[Term])] = {
      input match {
        case Equal (t1, t2) =>
          val leftHS = NormalizePoly(Plus(t1,Neg(t2)),free)
          if (Degree(leftHS,free) >= 3) {
            (False, Nil) :: Nil
          } else
            leftHS match {

              case Times(a, Power(free, Number(s))) =>
                if (s == 0 || s == 2) {
                  (Greater(a, Number(0)), epsilon :: infinity :: Neg(infinity) :: Nil) :: Nil
                }
                else {
                  (Greater(a, Number(0)), epsilon :: Nil) ::
                    (Less(a, Number(0)), Neg(epsilon) :: Nil) :: Nil
                }
              case Plus(Times(a2, Power(free1, Number(b2))), Plus(Times(a1, Power(free2, Number(b1))), Times(a0, Power(free3, Number(b0)))))
                if (free1 == free && free2 == free & free3 == free) =>
                val discriminant = Minus(Times(a1, a1), Times(Number(4), Times(a2, a0)))
                val Rpart = Divide(Power(discriminant, Number(0.5)), Times(Number(2), a2))
                val Lpart = Divide(Neg(a1), Times(Number(2), a2))
                val root1 = Plus(Rpart, Lpart)
                val root2 = Minus(Rpart, Lpart)
                val requirement = GreaterEqual(discriminant, Number(0))
                (requirement, root1 :: root2 :: Nil) :: Nil

              case Plus(Times(a1, Power(free1, Number(b1))), Times(a0, Power(free2, Number(b0))))
                if (free1 == free && free2 == free) =>
                val requirement = NotEqual(a1, Number(0))
                val root = Divide(Neg(a0), a1)
                (requirement, root :: Nil) :: Nil

              case _ => (False, Nil) :: Nil
            }
        case Less(t1,t2) =>
          val leftHS = NormalizePoly(Plus(t1,Neg(t2)), free)
          if (Degree(leftHS,free) >= 3)
            {
              (False, Nil) :: Nil
            } else
            leftHS match {
              case Times(a, Power(free, Number(s))) =>
                if (s == 0 || s == 2) {
                  (Greater(a,Number(0)), infinity :: Neg(infinity) :: Nil) :: Nil
                }
                else {
                  (Greater(a, Number(0)), epsilon :: Nil) ::
                    (Less(a, Number(0)), Neg(epsilon) :: Nil) :: Nil
                }
              case Plus(Times(a1, Power(free1, Number(b1))), Times(a0, Power(free2, Number(b0))))
                if (free1 == free && free2 == free) =>
                val requirement = NotEqual(a1, Number(0))
                val root1 = Plus(Divide(Neg(a0), a1), epsilon)
                val root2 = Plus(Divide(Neg(a0), a1), Neg(epsilon))
                (requirement, root1 :: Nil) :: (requirement, root2 :: Nil) :: Nil

              case Plus(Times(a2, Power(free1, Number(b2))), Plus(Times(a1, Power(free2, Number(b1))), Times(a0, Power(free3, Number(b0)))))
                if (free1 == free && free2 == free & free3 == free) =>
                val discriminant = Minus(Times(a1, a1), Times(Number(4), Times(a2, a0)))
                val Rpart = Divide(Power(discriminant, Number(0.5)), Times(Number(2), a2))
                val Lpart = Divide(Neg(a1), Times(Number(2), a2))
                val root11 = Plus(Rpart, Lpart)
                val root21 = Plus(Rpart, Neg(Lpart))
                val root12 = Plus(root11, epsilon)
                val root22 = Plus(root21, Neg(epsilon))
                val root13 = Plus(root11, Neg(epsilon))
                val root23 = Plus(root11, epsilon)
                val requirement = GreaterEqual(discriminant, Number(0))
                (requirement, root12 :: root22 :: Nil) ::
                  (requirement, root13 :: root23 :: Nil) :: Nil
               case _ => (False, Nil) :: Nil
            }
              case Greater(t1,t2) =>
                val leftHS = NormalizePoly(Plus(t1,Neg(t2)), free)
                if (Degree(leftHS,free) >= 3)
                {
                  (False, Nil) :: Nil
                } else
                  leftHS match {
                    case Times(a, Power(free, Number(s))) =>
                      if(s == 0 || s == 2) {
                        (Greater(a, Number(0)),epsilon :: Nil) :: Nil
                      }
                      else{
                        (Greater(a,Number(0)),epsilon :: Nil) ::
                          (Less(a,Number(0)), Neg(epsilon) :: Nil) :: Nil
                      }
                    case Plus(Times(a1, Power(free1, Number(b1))), Times(a0, Power(free2, Number(b0))))
                      if (free1 == free && free2 == free) =>
                      val requirement = NotEqual(a1, Number(0))
                      val root1 = Plus(Divide(Neg(a0), a1),epsilon)
                      val root2 = Plus(Divide(Neg(a0), a1),Neg(epsilon))
                      (requirement, root1 :: Nil) :: (requirement, root2 :: Nil) :: Nil

                    case Plus(Times(a2, Power(free1, Number(b2))), Plus(Times(a1, Power(free2, Number(b1))), Times(a0, Power(free3, Number(b0)))))
                      if (free1 == free && free2 == free & free3 == free) =>
                      val discriminant = Minus(Times(a1, a1), Times(Number(4), Times(a2, a0)))
                      val Rpart = Divide(Power(discriminant, Number(0.5)), Times(Number(2), a2))
                      val Lpart = Divide(Neg(a1), Times(Number(2), a2))
                      val root11 = Plus(Rpart, Lpart)
                      val root21 = Plus(Rpart, Neg(Lpart))
                      val root12 = Plus(root11,epsilon)
                      val root22 = Plus(root21,Neg(epsilon))
                      val root13 = Plus(root11,Neg(epsilon))
                      val root23 = Plus(root11,epsilon)
                      val requirement = GreaterEqual(discriminant, Number(0))
                      (requirement, root12 :: root22 :: Nil) ::
                        (requirement, root13 :: root23 :: Nil) :: Nil

                    case _ => (False,Nil) :: Nil
                  }
              case GreaterEqual(t1,t2) => Replacement(Greater(t1,t2),free) ::: Replacement(Equal(t1,t2),free)
              case LessEqual(t1,t2) => Replacement(Less(t1,t2),free) ::: Replacement(Equal(t1,t2),free)
              case NotEqual(t1,t2) => Replacement(Less(t1,t2),free) ::: Replacement(Greater(t1,t2),free)
            }
      }


  /** ComputeReplacements will taken in a formula of normalized polynomials in free and compute the set of all replacement requirement pairs for free */
  def ComputeReplacements(form : Formula, free :  Variable) : List[(Formula,List[Term])] = {
    form match {
      case       Not(f) =>    ComputeReplacements(f,free)
      case   And(f1,f2) =>    ComputeReplacements(f1,free) ::: ComputeReplacements(f2,free)
      case    Or(f1,f2) =>    ComputeReplacements(f1,free) ::: ComputeReplacements(f2,free)
      case Imply(f1,f2) =>    ComputeReplacements(f1,free) ::: ComputeReplacements(f2,free)
      case Equiv(f1,f2) =>    ComputeReplacements(f1,free) ::: ComputeReplacements(f2,free)

      case  Exists(xs,f) =>   ComputeReplacements(f,free)
      case  Forall(xs,f) =>   ComputeReplacements(f,free)

      case Greater(t1,t2)      =>      Replacement(Greater(t1,t2),free)
      case GreaterEqual(t1,t2) =>      Replacement(GreaterEqual(t1,t2),free)
      case LessEqual(t1,t2)    =>      Replacement(LessEqual(t1,t2),free)
      case Less(t1,t2)         =>      Replacement(Less(t1,t2),free)
      case Equal(t1,t2)        =>      Replacement(Equal(t1,t2),free)
      case NotEqual(t1,t2)     =>      Replacement(NotEqual(t1,t2),free)

    }
  }
  /**Substitute will take in a normalized polynomial, a replacement rule and a symbol x, and replace x according to the rule */
  def SubstituteTerm (Poly : Term, replacement : Term, free :  Variable) : Term = {
    Poly match {
      case Times(a,Power(free,exp)) => Times(a,Power(replacement,exp))
      case Plus(a,b) => Plus(SubstituteTerm(a,replacement,free),SubstituteTerm(b,replacement,free))
      case Times(a,b) => Times(SubstituteTerm(a,replacement,free),SubstituteTerm(b,replacement,free))
      case _ => Poly
    }
  }
  /** Substitute Form will take in a formula, a replacement rule and a symbol x and replace x according to the rule */
  def SubstituteForm (Form : Formula, replacement : Term, free :  Variable) : Formula = {
    Form match {
      case       Not(f) =>   Not(SubstituteForm(f,replacement,free))
      case   And(f1,f2) =>   And(SubstituteForm(f1,replacement,free),SubstituteForm(f2,replacement,free))
      case    Or(f1,f2) =>    Or(SubstituteForm(f1,replacement,free),SubstituteForm(f2,replacement,free))
      case Imply(f1,f2) => Imply(SubstituteForm(f1,replacement,free),SubstituteForm(f2,replacement,free))
      case Equiv(f1,f2) => Equiv(SubstituteForm(f1,replacement,free),SubstituteForm(f2,replacement,free))

      case  Exists(xs,f) => Exists(xs,SubstituteForm(f,replacement,free))
      case  Forall(xs,f) => Exists(xs,SubstituteForm(f,replacement,free))

      case Greater(t1,t2)      =>      Greater (SubstituteTerm(t1,replacement,free),SubstituteTerm(t2,replacement,free))
      case GreaterEqual(t1,t2) => GreaterEqual (SubstituteTerm(t1,replacement,free),SubstituteTerm(t2,replacement,free))
      case LessEqual(t1,t2)    =>    LessEqual (SubstituteTerm(t1,replacement,free),SubstituteTerm(t2,replacement,free))
      case Less(t1,t2)         =>         Less (SubstituteTerm(t1,replacement,free),SubstituteTerm(t2,replacement,free))
      case Equal(t1,t2)        =>        Equal (SubstituteTerm(t1,replacement,free),SubstituteTerm(t2,replacement,free))
      case NotEqual(t1,t2)     =>     NotEqual (SubstituteTerm(t1,replacement,free),SubstituteTerm(t2,replacement,free))

    }
  }
  /**
  def VirtualSubstitue(Input : Formula) : Formula =
  Input match {
    case Exists(x,f) =>
      val freevar : Variable = x

    case Forall(x,f) => False
  }
  /** SimplifySQ takes in a term of the form ((a + b*c^{1/2})/d)^k, where k >= 0 and simplifies it to be
    * 1 or (a' + b'*c'^{1/2})/d'
     */
  def SimplifySQ (poly : Term) : Term = {
    poly match {
      case Power(inside,Number(s)) =>
        inside match {
          case Plus(lpart,rpart) =>
            (lpart,rpart) match {
              case(Divide(Power(discriminant, Number(0.5)), Times(Number(2), a2)),)

            }
        }

    }
  }
  /** ModifySquareRoots takes in a normalizedPolynomial in x that's had a quadratic root substituted into and
    * simplifies the squareroot expression */
  def Reprocess (poly: Term) : Term = {
    poly match {
      case Plus(Times(a,Power(sub1,Number(s))),b) =>
        sub1 match
      case (Times(a,Power(sub2,Number(s)))) =>
    }
  }

    */
}
